module Board where

import Array
import IO
import Maybe
import Monad

import ST

import IOExts
import Word
import Bits

import Foreign.Marshal.Alloc
import Foreign.Marshal.Array
import Foreign.Ptr
import Foreign.Storable

data Tile
  = Space
  | Home
  | Wall
  |Water
    deriving (Show,Eq)

char2tile '.' = Space
char2tile '@' = Home
char2tile '#' = Wall
char2tile '~' = Water

tile2char Space = '.'
tile2char Home  = '@'
tile2char Wall  = '#'
tile2char Water = '~'

{-
tile2char Space = ''
tile2char Home  = '@'
tile2char Wall  = '#'
tile2char Water = 'x'
-}

char2word :: Char -> Word8
char2word '.' = 0
char2word '@' = 1
char2word '#' = 2
char2word '~' = 3

-- bit 1&2  type
-- bit 3    near water
-- bit 4    visited
-- bit 5    robot

word2tile 0 = Space
word2tile 1 = Home
word2tile 2 = Wall
word2tile 3 = Water

type CoordX = Int
type CoordY = Int
type Square = (CoordX,CoordY)

data Board = Board Int Int (Ptr Word8)

foreign import ccall "water_info" c_water_info :: Ptr Word8
					       -> Int
					       -> Int
					       -> IO ()

read_board :: Handle -> IO Board
read_board handle = do
  str <- hGetLine handle
  let [w,h] = map read (words str) :: [Int]
  rows <- sequence $ replicate h (hGetLine handle)
  ptr <- newArray [char2word c| r <- rows, c <- r]
  c_water_info ptr w h
  return $ Board w h ptr

update_board_remove_home :: Board -> Square -> IO ()
update_board_remove_home (Board w h ptr) (x,y) = do
  old <- peekByteOff ptr ((x-1)+(y-1)*w)
  pokeByteOff ptr ((x-1)+(y-1)*w) ( (old::Word8) .&. (complement 0x03))
  return ()

update_board_visited :: Board -> Square -> IO ()
update_board_visited (Board w h ptr) (x,y) = do
  old <- peekByteOff ptr ((x-1)+(y-1)*w)
  pokeByteOff ptr ((x-1)+(y-1)*w) ( (old::Word8) .|. 0x08)
  return ()

board_visited :: Board -> Square -> IO Bool
board_visited (Board w h ptr) (x,y) = do
  old <- peekByteOff ptr ((x-1)+(y-1)*w)
  return $ ((old::Word8) .&. 0x08) /= 0

update_board_robot_set:: Board -> Square -> IO ()
update_board_robot_set (Board w h ptr) (x,y) = do
  old <- peekByteOff ptr ((x-1)+(y-1)*w)
  pokeByteOff ptr ((x-1)+(y-1)*w) ( (old::Word8) .|. 0x10)
  return ()

update_board_robot_clear:: Board -> Square -> IO ()
update_board_robot_clear (Board w h ptr) (x,y) = do
  old <- peekByteOff ptr ((x-1)+(y-1)*w)
  pokeByteOff ptr ((x-1)+(y-1)*w) ( (old::Word8) .&. (complement 0x10))
  return ()

board_get_real (x,y) (Board w h ptr)
    | x >= 1 && x <= w &&
      y >= 1 && y <= h = f
    | otherwise = error "Outside board"
    where f = do w <- peekByteOff ptr ((x-1)+(y-1)*w)
		 return $ word2tile ( (w::Word8) .&. 3)

board_get_water_warning (x,y) (Board w h ptr)
    | x >= 1 && x <= w &&
      y >= 1 && y <= h = f
    | otherwise = error "Outside board"
    where f = do w <- peekByteOff ptr ((x-1)+(y-1)*w)
		 return $ ((w::Word8) .&. 0x4) /= 0

can_go_to b@(Board w h _) (x,y) = if x >= 1 && x <= w 
				  && y >= 1 && y <= h
				     then do t <- board_get_real (x,y) b
					     return $ t `elem` [Space,Home]
				     else return False

square_is tile b@(Board w h _) (x,y) = if x >= 1 && x <= w 
				       && y >= 1 && y <= h
					  then do t <- board_get_real (x,y) b
						  return $ t == tile
					  else return False

mk_board rows = do
  ptr <- newArray [char2word c| r <- reverse rows, c <- r]
  c_water_info ptr w h
  return $ Board w h ptr
	 where w = length (head rows)
	       h = length rows


b1 = mk_board $ replicate 300 $ replicate 300 '.'
b2 = mk_board $ replicate 1000 $ replicate 1000 '.'
b3 = mk_board ["..."
	      ,"#.."
	      ,"..."]

b4 = mk_board  ["..@...."
	       ,"......."
	       ,"##.~~~~"
	       ,"...~~~~"
	       ,"......."]

b5 = mk_board rows
    where rows = [".....................................................###########..............."
		 ,"..............................................................................."
		 ,"..........................................................@...................."
		 ,"..................~~~~~~~...........#................###########..............."
		 ,"..................~~~~~~~...........#........~~~~~~~~~~~~~....................."
		 ,"......................~~~...........#.......~~~~~~~~~~~~~~....................."
		 ,"....................................#..............~~~~~~......................"
		 ,"....~~~~~~~~~~~.....................#..............~~~~~~......................"
		 ,"....~~~~~~~~~~~.@...................#.........................................."
		 ,"....~~~~~~~~~~~.....................#.........................................."
		 ,"........................#.#.........######################################....."
		 ,"........................#.#.................................#.................."
		 ,"........................#.#.................................#.................."
		 ,"#########################.################################.###.################"
		 ,"........................#.#..................~~~~~~......#.....#..............."
		 ,".....~~~~~~~............#.#................~~~~~~~~~~....#.....#..............."
		 ,"...~~~~~~~...............................#~~~~~~~~~......###.###.....~~~~......"
		 ,".........................................#~~~~~~~...................~~~~~~~...."
		 ,".........................................#~~~~~~~....................~~~~~....."
		 ,".........................................#########....................~~~~....."
		 ,".......................................................................~~~....."
		 ,"..............................................................................."
		 ,"..............................................................................."
		 ,"..............................................................................."]
		 

b6 = mk_board [".~"
	      ,".."]

b7 = mk_board ["..~"
	      ,"..."]

b8 = mk_board ["..~"
	      ,"..."
	      ,"~.."]

b9 = mk_board ["...~"
	      ,"~..."
	      ,"...~"
	      ,"~..."]

b10 = mk_board [".................................................."
	       ,"................~~~~~............................."
	       ,"...............~~~~~~~~~.........................."
	       ,"................~~~~~~~~~........................."
	       ,"..................~~~~~~.........................."
	       ,".................................................."
	       ,".........#############...........#..........#....."
	       ,"...............#.................#..........#....."
	       ,"...............#.................#..........#....."
	       ,"...............#.................#..........#....."
	       ,"...............#.................############....."
	       ,"...............#.................#..........#....."
	       ,"...............#.................#..........#....."
	       ,"...............#.................#..........#....."
	       ,"...............#......~~~~.......#..........#....."
	       ,".....................~~~~~~......................."
	       ,"....~~~~~...........~~~~~~~~......................"
	       ,"...~~~~~~~.........~~~~~~~~~~....................."
	       ,"..~~~~~~~~~.......~~~~~~~~~~~~..........~~~~~~...."
	       ,"...~~~~~~........~~~~~~~~~~~~~~.........~~~~~~~..."
	       ,"....~~...........~~~~~~~~~~~~~~........~~~~~~....."
	       ,".................~~~~~~~~~~~~~...................."
	       ,".....................~~~~~........................"
	       ,".................................................."
	       ,"...............................#..#..............."
	       ,"#####..#########################..###..###########"
	       ,"~~~~~..~~~~~#..................#..#~~..#~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..................#..#~#..~~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..##############..#..#~~..#~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..#...............#..#~#..~~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..#...............#..#~~..#~~~~~~~~~~"
	       ,"~~~~~~.~~~~~#..#..##############..#~#..~~~~~~~~~~~"
	       ,"~~~~~.~~~~~~#..#..................#~~..#~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..#..................#~#..~~~~~~~~~~~"
	       ,"~~~~~..~~~~~#..####################~~..#~~~~~~~~~~" --11
	       ,"~~~~~..~~~~~#.....................#~#..~~~~~~~~~~~" --10
	       ,"~~~~~..~~~~~#.....................#~~..#~~~~~~~~~~" --9
	       ,"#####..#########################..#~#..###########" --8
	       ,"...............................#..###............." --7
	       ,".................................................." --6
	       ,".................................................." --5
	       ,"...................@.............................." --4
	       ,".................................................." --3
	       ,".................................................." --2
	       ,".................................................."]--1
