module Client where

import Array
import Char
import IO
import List
import Maybe
import Monad
import Socket
import System

import FiniteMap
import IOExts

import Ansi
import Board
import Config
import Distance
import Package
import Robot
import Util

get_line = hGetLine
put_line = hPutStrLn

{-
append_file name str = do
  h <- openFile name AppendMode
  hPutStr h str
  hClose h

get_line handle = do
  str <- hGetLine handle
  append_file "log.txt" (str ++ "\n")
  return str

put_line handle str = do
  hPutStrLn handle str
  append_file "log.txt" ("Send: " ++ str ++ "\n")
-}

type AI_Robot state = state -> BotInfo -> IO ((Bid,Command,String), state)

data BotInfo = BI
 {
   board        :: Board,
   rid          :: RobotID,
   max_weight   :: Int,
   fuel         :: Money,
   packages     :: [Package], -- at the current square
   robots       :: [Robot],
   pkg_place    :: FiniteMap PackageID Square,
   pkg_info     :: FiniteMap PackageID (Square, Weight), -- dest

   -- calculated once
   home         :: [Square],

   -- calculated after each move
   current_sq   :: Square,
   carry        :: [Package],
   free_weight  :: Int,
   dist_from_me :: Distance
--   dist_from_me :: Array Square Int
 }

get_pkg_weight bi p = w
    where Just (_,w) = lookupFM (pkg_info bi) p

get_pkg_dest bi p = sq
    where Just (sq,_) = lookupFM (pkg_info bi) p

show_board bi =
  let bb@(Board w h _) = board bi
      f sq = do t <- board_get_real sq bb
		warning <- board_get_water_warning sq bb
		f' sq [(tile2char t)] warning

      f' (x,y) map_str warning
	  | (x,y) == (current_sq bi)                 = return $ bg_red    ++ map_str ++ bg_default
	  | (x,y) `elem` [sq|(_,(sq,_))<-robots bi]  = return $ bg_green  ++ map_str ++ bg_default
--	  | warning && (map_str `elem` [".","@"])    = return $ bg_yellow ++ map_str ++ bg_default
	  | (x,y) `elem` [sq | (_,(sq,_))<-carry bi] = return $ bg_blue   ++ map_str ++ bg_default
	  | map_str == "~"                           = return $ bg_yellow ++ map_str ++ bg_default
	  | map_str == "#"                           = return $ bg_black  ++ map_str ++ bg_default
	  | (x,y) `elem` (map head $
			  group $
			  sort $
			  map snd $
			  fmToList $
			  pkg_place bi)              = return $ bg_cyan ++ map_str ++ bg_default
	  | otherwise                                = return $ map_str
      line y = do xs <- sequence [ f (x,y) | x <- [1..w]]
		  return $ concat xs
  in do sequence [line y >>= putStrLn' | y <- reverse [1..h]] 
	return ()

print_botinfo bi = do
  putStrLn' $"Home: " ++ show (home bi)
  putStrLn' $"ID:   " ++ show (rid bi)
  putStrLn' $"curr: " ++ show (current_sq bi)
  putStrLn' $"fuel: " ++ show (fuel bi)
  putStrLn' $"carry:" ++ show (carry bi)

update_fuel (a,b,c) x = (a,b,c-x)

update_pos (x,y) (X val) = (val,y)
update_pos (x,y) (Y val) = (x,val)
update_pos (x,y) N       = (x,y+1)
update_pos (x,y) S       = (x,y-1)
update_pos (x,y) E       = (x+1,y)
update_pos (x,y) W       = (x-1,y)

update_robot_list :: [ (RobotID, RobotInfo) ]
		  -> [ (RobotID,[Action]) ]
		  -> [ (RobotID, RobotInfo) ]
update_robot_list robots [] = robots
update_robot_list robots (x:rest) = update_robot_list (up robots (x)) rest
  where up [] (num,act) = [(num, foldl f ((0,0),[]) act)]
	up ((rid,x):rest) (num,act)
	   | rid == num = (rid, foldl f x act) : rest
	   | otherwise  = (rid,x) : up rest (num,act)

        f :: RobotInfo -> Action -> RobotInfo
	f ((x,y),pkgs) (X val) = ((val, y), pkgs)
	f ((x,y),pkgs) (Y val) = ((x, val), pkgs)
	f ((x,y),pkgs) W       = ((x-1, y), pkgs)
	f ((x,y),pkgs) E       = ((x+1, y), pkgs)
	f ((x,y),pkgs) N       = ((x, y+1), pkgs)
	f ((x,y),pkgs) S       = ((x, y-1), pkgs)
	f ((x,y),pkgs) (P v)   = ((x, y), v:pkgs)
	f ((x,y),pkgs) (D v)   = ((x, y), pkgs\\[v])

fromJust' (Just x) = x
fromJust' _ = error "fj"

update_pkg_place [] bi = bi
update_pkg_place ((rid,ac):rest) bi = update_pkg_place rest (foldr up bi ac)
  where up (P v) bi = bi { pkg_place = delFromFM (pkg_place bi) v }
	up (D v) bi = let i = lookupFM (pkg_info bi) v
			  (robot_sq,_) = fromJust $ lookup rid (robots bi)
		      in case i of
		           Nothing -> bi { pkg_place = addToFM (pkg_place bi) v robot_sq }
			   Just (dest,w) -> if dest == robot_sq && w <= (max_weight bi)
					       then bi
					       else bi { pkg_place = addToFM (pkg_place bi) v robot_sq }
	up _ bi = bi

putStr' x = putStr x
putStrLn' x = putStrLn (x ++ clear_eol)

get_homes b@(Board w h _) = do
   xs <- sequence [board_get_real (x,y) b >>= (\q -> return ((x,y),q))
		   | y<-[1..h], x<-[1..w]]
   return [ sq | (sq,t) <- xs, t == Home]

---- Arg, I want to rewrite this function, but I don't want to risk introducing bugs..
play_game :: Handle -> AI_Robot a -> a -> Board -> Me -> [Robot] -> IO ()
play_game handle prg state board@(Board width height _) me@(me_id,me_max,me_fuel) rbots = do
  putStr' cursor_off
  di1 <- new_distance board
  h <- get_homes board
  let bi = BI {
		  board = board,
		  home = h,
--		  dist_from_me = distance_to board (current_sq bi),
		  dist_from_me = di1,
		  current_sq = fst $ fromJust $ lookup (rid bi) (robots bi),
		  carry = [],
		  rid = me_id,
		  max_weight = me_max,
		  free_weight = me_max,
		  fuel = me_fuel,
		  packages = [],
		  robots = rbots,
		  pkg_place = emptyFM,
		  pkg_info = emptyFM
                }
  update_distance_from (dist_from_me bi) (current_sq bi)
  sequence [update_board_robot_set board sq | (_,(sq,_)) <- rbots]
  bi_ref <- newIORef bi
  pg state bi_ref
  return ()

  where
        pg state bi_ref = do
	  bi <- readIORef bi_ref

--          putStr (clear_screen)
          putStr' (goto_topleft)
	  
          when (width < 200 && height < 60) (show_board bi)

	  str' <- get_line handle

          -- Bhhaaaa! They changed the server to send extra strings...
          -- Why did you have to do that???
          -- This functions was ugly enough before...
	  str <- if str' /= "" && not (isDigit (head str'))
		   then do putStrLn' $ "Game over"
			   sequence $ repeat (get_line handle >>= putStrLn')
			   return ""
		   else return str'

	  --putStrLn' $"input (pkg): " ++ str
	  let pkgs = parse_packages str

	  putStrLn' $ "Fuel    : " ++ (show $ fuel bi)
	  putStrLn' $ "Pos     : " ++ (show $ current_sq bi)
	  putStrLn' $ "Home    : " ++ (show $ length (home bi))
--	  putStrLn' $ "Robots  : " ++ (show $ robots bi)
--	  putStrLn' $ "Packages: " ++ (show $ length pkgs)	  
--	  putStrLn' $ "Packages: " ++ (show $ pkgs)	  
	  putStrLn' $ "Carry   : " ++ (show $ carry bi)
	  putStrLn' $ "Free    : " ++ (show $ free_weight bi)
	  putStrLn' $ "pkg     : " ++ (show $ map (\x->(head x,length x)) $ group $
				       sort $ map snd $ fmToList $ pkg_place bi)

          -- we only care about packages we can lift
          let pkgs' = [p | p@(_,(_,w)) <- pkgs, w <= (max_weight bi)]

          -- remove everything we believe about this tile

          modifyIORef bi_ref (\bi -> bi {pkg_place = filterFM
					    (\k v->v /= (current_sq bi)) (pkg_place bi) })

          modifyIORef bi_ref (\bi -> bi {pkg_place = addListToFM (pkg_place bi) 
					                [(p,current_sq bi)| (p,_) <- pkgs']})
          
          modifyIORef bi_ref (\bi -> bi {pkg_info = addListToFM (pkg_info bi) pkgs,
					 packages = pkgs})
	  bi <- readIORef bi_ref

	  ((b,c,info),st) <- prg state bi

	  put_line handle (bid_cmd2str (b,c))

	  putStrLn' $ "My move : " ++ (bid_cmd2str (b,c)) ++ " (" ++ info ++ ")"
		  
          str' <- get_line handle
	  --putStrLn' $"input (act): " ++ str

          -- Bhhaaaa! They changed the server to send extra strings...
	  str <- if str' /= "" && head str' /= '#'
		   then do putStrLn' $ "Game over"
			   sequence $ repeat (get_line handle >>= putStrLn')
			   return ""
		   else return str'

	  let actions = parse_robot_actions str

--	  putStrLn' $ "Actions : " ++ (show actions)
	  putStrLn' ""

	  sequence [update_board_robot_clear board sq | (_,(sq,_)) <- (robots bi)]

          modifyIORef bi_ref (\bi -> bi {fuel = (fuel bi) - (abs b)})
          modifyIORef bi_ref (\bi -> bi {robots = update_robot_list (robots bi) actions})
          modifyIORef bi_ref (\bi -> bi {robots = [ r | r@(i,_)<-robots bi, i `elem` (map fst actions)]})
          modifyIORef bi_ref (\bi -> bi {current_sq = fst $ fromJust $ lookup (rid bi) (robots bi)})
          modifyIORef bi_ref (\bi -> bi {carry      = 
					 [ (d,i)
					   | d <- snd $ fromJust $ lookup (rid bi) (robots bi)
					   , let Just i = lookupFM (pkg_info bi) d ]})
          modifyIORef bi_ref (\bi -> bi {free_weight = (max_weight bi) - sum [w | (_,(_,w)) <- carry bi]})

 	  bi <- readIORef bi_ref

	  sequence [update_board_robot_clear board sq | (_,(sq,_)) <- (robots bi)]

	  update_distance_from (dist_from_me bi) (current_sq bi)
	  when (current_sq bi `elem` home bi)
		   (do modifyIORef bi_ref (\bi -> bi {home = (home bi) \\ [current_sq bi]})
		       {- update_board_remove_home board (current_sq bi) -})

          update_board_visited board (current_sq bi)

          modifyIORef bi_ref (update_pkg_place actions)

          putStr' clear_eos
	  --system "sleep 1"
	  pg st bi_ref
	  
          `catch` (\s->putStrLn' "" >> putStrLn' ("Aaarg, I'm dead" ++ cursor_on) )

start_bot_server :: String -> PortNumber -> AI_Robot a -> a -> IO ()
start_bot_server host port bot state = do
   handle <- connectTo host (PortNumber port)
   hSetBuffering handle NoBuffering
   putStrLn "Connected"
   hPutStrLn handle "Player"
   b <- read_board handle
   (m,r) <- read_config handle
   print m
   putStr clear_screen
   play_game handle bot state b m r

start_bot :: AI_Robot a -> a -> IO ()
start_bot bot state = do
  [host,port] <- getArgs
  start_bot_server host (toEnum (read port)) bot state
